/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

// compute basic proposal distribution for a particle from the motion
// model

// this motion model isn't the greatest.  we're relying on odometry
// data consisting of a sequence of poses.  we can compute a control
// input of the form u=[d alpha]' from the data, but it's not very
// easy to get the sign of d correct, so if the robot is ever moving
// in reverse that approach tends to blow up.  instead we'll go with
// this simpler approach, which assumes errors in x,y,theta are
// independent.  luckily rbpf is pretty robust to this and it works
// quite well most of the time.  (and the improved proposal makes it
// even less of an issue.)

#include "slam.hpp"

void motion_model
  (const pose_t &delta, const pose_t &ref_frame, const pose_t &particle_frame,
   pose_t &mu, cov3_t &cov)
{
  const double sint = sin(particle_frame.t() - ref_frame.t());
  const double cost = cos(particle_frame.t() - ref_frame.t());
  mu = pose_t(particle_frame.x() + cost*delta.x() - sint*delta.y(),
	      particle_frame.y() + sint*delta.x() + cost*delta.y(),
	      particle_frame.t() + delta.t());
  const pose_t mydelta(mu - particle_frame);
  cov[0][0] += param::err_trans*mydelta.x() * param::err_trans*mydelta.x();
  cov[1][1] += param::err_trans*mydelta.y() * param::err_trans*mydelta.y();
  cov[2][2] += param::err_norm*param::err_norm * (mydelta.x()*mydelta.x()+mydelta.y()*mydelta.y()) +
    param::err_rot*mydelta.t() * param::err_rot*mydelta.t();
}
