/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

// THIS CODE IS BEING REPLACED BY rbpf_block2

// RBPF SLAM with block proposal distribution

////////////////////////////////////////////////////////////////////////
// a lot of this code is nearly identical to parts of rbpf_improved,
// but different enough that i couldn't just call rbpf_improved.  so,
// if you make a change to rbpf_improved, you should check carefully
// here to see if you need to make similar changes.
////////////////////////////////////////////////////////////////////////

#include "rbpf.hpp"


void slam_rbpf_block::add_features
  (const std::vector<feature_type> &features,
   const std::vector<uint32_t> &associations,
   const pose_t &ref_frame)
{
  // save the data from this frame for later use in sampling
  fixed_lag_data d;
  d.u = last_delta;
  d.u_ref_frame = last_ref_frame;
  d.z = features;
  d.z_ref_frame = ref_frame;
  fixed_lag_push_front(lag_data, d);

  // either we know all associations or none
  assert(associations.size() == 0 || associations.size() == features.size());

  double W = 0;
  uint32_t i;

  std::vector<pose_model_t> pose_models(param::lag_time+2); // includes t-L
  double w;

  // iterate over all particles
  for(i = 0; i < particles.size(); ++i) {
    lag_particle_t &p = particles[i];

    // will be modified in filter_forward if data associations are unknown
    fixed_lag_push_front(p.n, associations);
  
    // forward filtering step
    int k = int(lag_data.size())-2;

    pose_model_t &start = pose_models[k+1];
    start.mu = *(p.poses.rbegin()); // pose at t-L;
    start.cov = cov_epsilon; // see rbpf.hpp

    map_type tmp_map = *(p.maps.rbegin()); // map from t-L
    std::list<fixed_lag_data>::reverse_iterator dri = lag_data.rbegin();
    std::list< std::vector<uint32_t> >::reverse_iterator ni = p.n.rbegin();
    for(++dri, ++ni; k >= 0; --k, ++dri, ++ni) {
      w = filter_forward(*dri, pose_models[k+1], pose_models[k], tmp_map, *ni);

      assert(dri->z.size() == ni->size());
      printf("forward %d: ", k);
      std::cout << pose_models[k].mu << std::endl << pose_models[k].cov << std::endl;


    }

    // FIXME: needs more bookkeeping for the block proposal
    // update observation counts for spurious landmark removal using
    // the current observations
    //    for(uint32_t a = 0; a < p.n.front().size(); ++a)
    //      if(p.n.front()[a] != 0)
    //	      ++p.obs_count[tmp_map.find_id(p.n.front()[a])];

    // update particle weight; w is now the likelihood of the current
    // measurement (z_t) for the forward-filtered pose/map
    // distribution
    p.w *= w;

    if(isnan(p.w))
      p.w = 0;

    // backward sampling step
    p.pose = gaussian(pose_models[0].mu, pose_models[0].cov); // directly sample current time step


    printf("backward 0: ");
    std::cout << pose_models[0].mu << std::endl << pose_models[0].cov << std::endl;


    p.history.push_back(p.pose);
    p.poses.clear();
    p.poses.push_back(p.pose);
    std::list<fixed_lag_data>::iterator di = lag_data.begin();
    for(k = 1; k < int(lag_data.size())-1; ++k, ++di) {
      // compute the backward model from u_{t-k+1} and x_{t-k+1} and
      // merge it with the forward-filtered model
      pose_t mu_b(-di->u.x(), -di->u.y(), -di->u.t()); cov3_t cov_b = cov_epsilon;
      motion_model(mu_b, di->u_ref_frame, p.poses.back(), mu_b, cov_b);
      merge(pose_models[k].mu, pose_models[k].cov, mu_b, cov_b);

      // sample x_{t-k}
      p.poses.push_back(gaussian(pose_models[k].mu, pose_models[k].cov));

      printf("backward %d: ", k);
      std::cout << pose_models[k].mu << std::endl << pose_models[k].cov << std::endl;

    }

    printf("lag_data.size = %d\n", lag_data.size());

    if(int(lag_data.size())-1 < int(param::lag_time)) {
      p.poses.push_back(start.mu); // haven't gone long enough to discard this yet
      printf("added start\n");
    }


    std::list<pose_t>::iterator blah=p.poses.begin();
    printf("sampled poses: ");
    for(; blah != p.poses.end(); ++blah)
      printf("[%g %g %g] ", blah->x(), blah->y(), blah->t());
    printf("\ncurrent pose: %g %g %g\n", p.pose.x(), p.pose.y(), p.pose.t());
    if(p.poses.size() > 1) {
      blah = p.poses.begin(); ++blah;
      if(fabs(p.pose.t()-blah->t()) > 0.4) {
	printf("wha?\n"); assert(false);
      }
    }



    // update the map conditioned on the newly drawn set of poses
    tmp_map = *(p.maps.rbegin());
    p.maps.clear();
    dri = lag_data.rbegin();
    std::list<pose_t>::reverse_iterator pri = p.poses.rbegin();
    ni = p.n.rbegin();
    if(int(lag_data.size())-1 < int(param::lag_time)) {
      p.maps.push_front(tmp_map);
      ++pri; // see above where we inserted this pose
    }
    // note that we aren't recomputing data associations here (yet
    // again) but that it might be a good idea to do so
    //    for(++dri, ++ni; pri != p.poses.rend(); ++dri, ++pri, ++ni) {
    for(++dri, ++ni, k = lag_data.size()-1; pri != p.poses.rend(); ++dri, ++pri, ++ni, --k) {
      // transform features to the sampled frame of this particle
      cur_features.resize(dri->z.size());
      for(uint32_t f = 0; f < dri->z.size(); ++f)
	feature_to_particle_frame
	  (dri->z[f], dri->z_ref_frame, cur_features[f], *pri);

      // update WITHOUT inflated uncertainty (unlike rbpf_improved)
      update_map_improved(tmp_map, cur_features, *ni, *pri, w);
      p.maps.push_front(tmp_map);

      //      double asdf;
      //      update_map_improved(tmp_map, cur_features, *ni, *pri, asdf, &(pose_models[k].mu), &(pose_models[k].cov));
      //      p.maps.push_front(tmp_map);
    }

    p.history.fixed_lag_modify(p.poses);
    W += p.w;

    // FIXME: this requires more bookkeeping with the block proposal
    // remove landmarks we should have seen but didn't
    // remove_spurious_landmarks(p.pose, p.get_map(), in_range, all_points, p.obs_count);
  }

  // this is kind of a catastrophic failure, but do what we can
  if(W == 0) {
    for(i = 0; i < particles.size(); ++i)
      particles[i].w = 1.0;
    W = double(particles.size());
  }

  // normalize importance weights and compute "best" particle
  best = 0;
  for(i = 0; i < particles.size(); ++i) {
    particles[i].w /= W;
    if(particles[i].w > particles[best].w)
      best = i;
  }

  Neff = neff(particles);

  // check if we need to resample, and if so, do it
  if(Neff < param::resample_neff_percent * particles.size())
    resample(particles);

  // we've now done a SLAM update up to the most recent control input
  new_odometry = true;

  if(param::gui) {
    // for rendering later
    frame = ref_frame;
    cur_features = features;
  }

  do_analysis();
}


// this version of forward filtering does not implement a full EKF,
// since it assumes landmarks are independent of each other and
// updates the landmark distributions individually.  a better (but
// more computationally intensive) approach is to use a full EKF on
// the combined state vector [x^r x^m]'.  for some details, see the
// tech report.

// returns likelihood of the measurements from this time step using
// the forward-filtered pose/map distribution
double slam_rbpf_block::filter_forward
  (const fixed_lag_data &d, const pose_model_t &last,
   pose_model_t &cur, map_type &tmp_map, std::vector<uint32_t> &associations)
{
  // forward filtering as implemented here is basically rbpf_improved
  // without actually picking sampled values for the poses

  // prediction step
  cur.cov = cov_epsilon;
  motion_model(d.u, d.u_ref_frame, last.mu, cur.mu, cur.cov);
  cur.cov = cur.cov + last.cov;

  std::cout << "after motion_model: " << cur.mu << std::endl << cur.cov << std::endl;

  // get in-range landmarks
  std::vector<map_type::iterator> in_range;
  if(associations.size() != d.z.size())
    tmp_map.in_range(cur.mu, param::max_range, in_range);

  uint32_t f;
  double mahalanobis_dist;
  feature_type feat;
  map_type::iterator assoc;
  for(uint32_t f = 0; f < d.z.size(); ++f) {
    // incorporate measurements in the proposal
    feature_to_particle_frame(d.z[f], d.z_ref_frame, feat, cur.mu);
    if(associations.size() == d.z.size()) // already have a data association
      assoc = tmp_map.find_id(associations[f]);
    else {
      if(in_range.size()) {
	// do data association (inflate uncertainty by current proposal)
	data_association(add_pose_uncertainty(feat, cur.mu, cur.cov), in_range, assoc, mahalanobis_dist);
	if(mahalanobis_dist >= param::da_ml_threshold)
	  assoc = 0;
      } else assoc = 0;
    }
    if(assoc) {

      std::cout << "before: " << cur.mu << std::endl << cur.cov << std::endl
		<< "feature: " << feat << std::endl << feat.cov << std::endl
		<< "landmark: " << assoc->landmark << std::endl << assoc->landmark.cov
		<< assoc->id << std::endl;

      improve_proposal(cur.mu, cur.cov, feat, assoc->landmark);

      std::cout << "after: " << cur.mu << std::endl << cur.cov << std::endl;
    }
  }

  // transform features to the proposal mean frame
  cur_features.resize(d.z.size());
  for(f = 0; f < d.z.size(); ++f)
    feature_to_particle_frame
      (d.z[f], d.z_ref_frame, cur_features[f], cur.mu);

  // if necessary do data association conditioned on the current proposal
  if(associations.size() != d.z.size()) {
    associations.resize(d.z.size());
    for(f = 0; f < cur_features.size(); ++f) {
      if(in_range.size())
	data_association(cur_features[f], in_range, assoc, mahalanobis_dist);
      if(!in_range.size() || mahalanobis_dist >= param::da_ml_threshold)
	associations[f] = 0; // no association, we'll add a new landmark
      else
	associations[f] = assoc->id;
    }
  }

  // update map with the current observations and correspondences
  double w = 1; // must be 1: update_map_improved multiplies by the initial value
  update_map_improved
    (tmp_map, cur_features, associations, cur.mu, w, &cur.mu, &cur.cov);

  return w;
}
