/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

// THIS CODE NEEDS WORK.  There are still some bugs that I haven't
// fixed.

#include "rbpf.hpp"

slam_rbpf_block::slam_rbpf_block(const pose_t &start_pose)
  : slam_rbpf_improved<block_particle_t>(start_pose)
{
  map_type m;
  std::vector<uint32_t> n(0);
  fixed_lag_data blank;
  blank.u = pose_t(0,0,0);
  blank.u_ref_frame = blank.z_ref_frame = pose_t(0,0,0);
  for(uint32_t i = 0; i < param::N; ++i) {
    block_particle_t &p = particles[i];
    fixed_lag_push_front(p.poses, start_pose);
    fixed_lag_push_front(p.maps, m);
    fixed_lag_push_front(p.n, n);
  }
  fixed_lag_push_front(lag_data, blank);
}

void slam_rbpf_block::add_features
  (const std::vector<feature_type> &features,
   const std::vector<uint32_t> &associations,
   const pose_t &ref_frame)
{
  // save the data from this frame for later use in sampling
  fixed_lag_data d;
  d.u = last_delta;
  d.u_ref_frame = last_ref_frame;
  d.z = features;
  d.z_ref_frame = ref_frame;
  fixed_lag_push_front(lag_data, d);

  printf("ldsize = %d\n", lag_data.size());
  for(int q=0; q < lag_data.size(); ++q)
    printf("%d [%g %g %g] %d feat\n", q,
	   lag_data[q].u.x(), lag_data[q].u.y(), lag_data[q].u.t(),
	   lag_data[q].z.size());


  // either we know all associations or none
  assert(associations.size() == 0 || associations.size() == features.size());

  double W = 0;
  uint32_t i;

  std::vector<pose_model_t> pose_models(param::lag_time+2); // includes t-L
  double w;

  // iterate over all particles
  for(i = 0; i < particles.size(); ++i) {
    block_particle_t &p = particles[i];

    // will be modified in filter_forward if data associations are unknown
    fixed_lag_push_front(p.n, associations);

    int k = int(p.poses.size()) - 1;

    pose_model_t &start = pose_models[k+1];
    start.mu = p.poses[k]; // pose at t-L
    start.cov = cov_epsilon;
    map_type tmp_map = p.maps[k]; // map from t-L

#if 0
    printf("start map for forward filtering: %d\n", k);
#endif
    printf("poses: %d\n", p.poses.size());
    for(int q = 0; q < p.poses.size(); ++q)
      printf("%d [%g %g %g]\n", q, p.poses[q].x(), p.poses[q].y(), p.poses[q].t());
    printf("maps: %d\n", p.maps.size());
    for(int q = 0; q < p.maps.size(); ++q)
      printf("%d %d\n", q, p.maps[q].size());
    printf("assoc: %d\n", p.n.size());
    for(int q = 0; q < p.n.size(); ++q)
      printf("%d %d\n", q, p.n[q].size());
    printf("k = %d\n\n", k);

    // filter forward starting at t-L+1 and going to t (k == lag time)
    for(; k >= 0; --k)
      w = filter_forward(lag_data[k], pose_models[k+1], pose_models[k], tmp_map, p.n[k]);

    // now we have the forward filtered distributions for each of the
    // poses

    // update particle weight; w is now the likelihood of the current
    // measurement (z_t) for the forward-filtered pose/map
    // distribution
    p.w *= w;
    if(isnan(p.w))
      p.w = 0;

#if 0
    printf("forward:\n");
    for(k = 0; k < int(pose_models.size()); ++k)
      std::cout << k << ": " << lag_data[k].u << std::endl
		<< pose_models[k].mu << std::endl << pose_models[k].cov << std::endl;
#endif

    // backward sampling step
    p.pose = gaussian(pose_models[0].mu, pose_models[0].cov); // directly sample current time step

    p.history.push_back(p.pose);
    p.poses.clear();
    p.poses.push_back(p.pose);

#if 0
    printf("backward:\n");
#endif

    for(k = 0; k < int(lag_data.size()) - 2; ++k) {
      // compute the backward model from u_{t-k+1} and x_{t-k+1} and
      // merge it with the forward-filtered model
      const fixed_lag_data &d = lag_data[k];
      pose_t mu_b(-d.u.x(), -d.u.y(), -d.u.t());
      cov3_t cov_b = cov_epsilon;

#if 0
      std::cout << k+1 << "->" << k << ": " << d.u << std::endl;
#endif

      motion_model(mu_b, d.u_ref_frame, p.poses.back(), mu_b, cov_b);

      // FIXME remove
      //      motion_model(mu_b, d.u_ref_frame, pose_models[k].mu, mu_b, cov_b);

#if 0
      std::cout << k+1 << ": " << d.u << std::endl << mu_b << std::endl << cov_b << std::endl;
#endif

      merge(pose_models[k+1].mu, pose_models[k+1].cov, mu_b, cov_b);

      // sample x_{t-k}
      p.poses.push_back(gaussian(pose_models[k+1].mu, pose_models[k+1].cov));
    }

#if 0
    printf("merged:\n");
    for(k = 0; k < int(pose_models.size()); ++k)
      std::cout << k << ": " << pose_models[k].mu << std::endl << pose_models[k].cov << std::endl;

    printf("sampled:\n");
    for(k = 0; k < int(p.poses.size()); ++k)
      std::cout << k << ": " << p.poses[k] << std::endl;
#endif

    // now update map conditioned on the newly drawn set of poses
    k = int(lag_data.size()) - 2;

    printf("start map = frame %d, maps = %d, poses = %d, n = %d, ld = %d\n",
	   k, p.maps.size(), p.poses.size(), p.n.size(), lag_data.size());
    printf("poses: %d\n", p.poses.size());
    for(int q = 0; q < p.poses.size(); ++q)
      printf("%d [%g %g %g]\n", q, p.poses[q].x(), p.poses[q].y(), p.poses[q].t());
    printf("maps: %d\n", p.maps.size());
    for(int q = 0; q < p.maps.size(); ++q)
      printf("%d %d\n", q, p.maps[q].size());
    printf("assoc: %d\n", p.n.size());
    for(int q = 0; q < p.n.size(); ++q)
      printf("%d %d\n", q, p.n[q].size());
    printf("k = %d\n=========================\n", k);

    tmp_map = p.maps[k];
    p.maps.clear();
    if(k < int(param::lag_time) - 1) { // can't discard these yet
      p.poses.push_back(start.mu);
      fixed_lag_push_front(p.maps, tmp_map);
    }

    // note that we aren't recomputing data associations here (yet
    // again) but that it might be a good idea to do so
    for(; k >= 0; --k) {
      // transform features to the sampled frame of this particle
      const fixed_lag_data &d = lag_data[k];
      cur_features.resize(d.z.size());
      for(uint32_t f = 0; f < d.z.size(); ++f)
	feature_to_particle_frame(d.z[f], d.z_ref_frame, cur_features[f], p.poses[k]);

#if 0
      printf("pose [%g %g %g] cur map %d, features %d, ",
	     p.poses[k].x(), p.poses[k].y(), p.poses[k].t(),
	     tmp_map.size(), cur_features.size());
#endif

      // update WITHOUT inflated uncertainty (unlike rbpf_improved)
      update_map_improved(tmp_map, cur_features, p.n[k], p.poses[k], w);
      fixed_lag_push_front(p.maps, tmp_map);

#if 0
      printf("new map %d\n", tmp_map.size());
#endif
    }

    p.history.fixed_lag_modify(p.poses);
    W += p.w;
  }

  // this is kind of a catastrophic failure, but do what we can
  if(W == 0) {
    for(i = 0; i < particles.size(); ++i)
      particles[i].w = 1.0;
    W = double(particles.size());
  }

  // normalize importance weights and compute "best" particle
  best = 0;
  for(i = 0; i < particles.size(); ++i) {
    particles[i].w /= W;
    if(particles[i].w > particles[best].w)
      best = i;
  }

  Neff = neff(particles);

  // check if we need to resample, and if so, do it
  if(Neff < param::resample_neff_percent * particles.size())
    resample(particles);

  // we've now done a SLAM update up to the most recent control input
  new_odometry = true;

  if(param::gui) {
    // for rendering later
    frame = ref_frame;
    cur_features = features;
  }

  do_analysis();
}

// this version of forward filtering does not implement a full EKF,
// since it assumes landmarks are independent of each other and
// updates the landmark distributions individually.  a better (but
// more computationally intensive) approach is to use a full EKF on
// the combined state vector [x^r x^m]'.  for some details, see the
// tech report.

// returns likelihood of the measurements from this time step using
// the forward-filtered pose/map distribution
double slam_rbpf_block::filter_forward
  (const fixed_lag_data &d, const pose_model_t &last,
   pose_model_t &cur, map_type &tmp_map, std::vector<uint32_t> &associations)
{
  // forward filtering as implemented here is basically rbpf_improved
  // without actually picking sampled values for the poses

  // prediction step
  cur.cov = last.cov;
  motion_model(d.u, d.u_ref_frame, last.mu, cur.mu, cur.cov);

  // get in-range landmarks
  std::vector<map_type::iterator> in_range;
  if(associations.size() != d.z.size())
    tmp_map.in_range(cur.mu, param::max_range, in_range);

  uint32_t f;
  double mahalanobis_dist;
  feature_type feat;
  map_type::iterator assoc;
  for(uint32_t f = 0; f < d.z.size(); ++f) {
    // incorporate measurements in the proposal
    feature_to_particle_frame(d.z[f], d.z_ref_frame, feat, cur.mu);
    if(associations.size() == d.z.size()) // already have a data association
      assoc = tmp_map.find_id(associations[f]);
    else {
      if(in_range.size()) {
	// do data association (inflate uncertainty by current proposal)
	data_association(add_pose_uncertainty(feat, cur.mu, cur.cov), in_range, assoc, mahalanobis_dist);
	if(mahalanobis_dist >= param::da_ml_threshold)
	  assoc = 0;
      } else assoc = 0;
    }
    if(assoc)
      improve_proposal(cur.mu, cur.cov, feat, assoc->landmark);
  }

  // transform features to the proposal mean frame
  cur_features.resize(d.z.size());
  for(f = 0; f < d.z.size(); ++f)
    feature_to_particle_frame(d.z[f], d.z_ref_frame, cur_features[f], cur.mu);

  // if necessary do data association conditioned on the current proposal
  if(associations.size() != d.z.size()) {
    associations.resize(d.z.size());
    for(f = 0; f < cur_features.size(); ++f) {
      if(in_range.size())
	data_association(cur_features[f], in_range, assoc, mahalanobis_dist);
      if(!in_range.size() || mahalanobis_dist >= param::da_ml_threshold)
	associations[f] = 0; // no association, we'll add a new landmark
      else
	associations[f] = assoc->id;
    }
  }

  // update map with the current observations and correspondences
  double w = 1; // must be 1: update_map_improved multiplies by the initial value
  update_map_improved
    (tmp_map, cur_features, associations, cur.mu, w, &cur.mu, &cur.cov);

  return w;
}
