/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

// THIS CODE NEEDS WORK.  There are still some bugs that I haven't
// fixed.


// Rao-Blackwellized constraint filter -- based on FS1.0 for
// simplicity but could certainly be modified to use the improved
// proposal distribution

#include "rbpf.hpp"

slam_rbpf_constrained::slam_rbpf_constrained(const pose_t &start_pose)
  : generic_slam_rbpf<constrained_particle_t>()
{
  // copied from rbpf.cpp
  for(uint32_t i = 0; i < param::N; ++i) {
    particles[i].pose = start_pose;
    particles[i].history.push_back(start_pose);
    particles[i].w = 1.0/param::N;
    particles[i].L_max = 0;
  }
}

// copied from rbpf.cpp
void slam_rbpf_constrained::add_odometry(const pose_t &delta, const pose_t &ref_frame)
{
  pose_t mu;
  cov3_t cov;
  // project all the particles forward
  for(uint32_t i = 0; i < particles.size(); ++i) {
    constrained_particle_t &p = particles[i];
    cov = cov_epsilon; // motion_model is additive wrt cov
    motion_model(delta, ref_frame, p.pose, mu, cov);
    p.pose = gaussian(mu, cov); // sample new pose
    p.history.push_back(p.pose); // update pose history
  }
}

// initialize a new landmark
void slam_rbpf_constrained::insert
  (constrained_particle_t &p, std::vector<map_type::iterator> &landmarks,
   const feature_type &feat, uint32_t id)
{
  map_type::iterator lm;
  if(id == 0)
    lm = p.map.insert(feat);
  else
    lm = p.map.insert(feat, id);

  p.backup[lm] = feat; // initialize backup state
  p.accum[lm] = feat;
  p.accum[lm].cov = cov_infinity; // initialize measurement accumulator

  p.L[lm] = p.L_max++; // set superlandmark id

  typedef constraint_type::param_type cp_type;
  constraint_type C; // constraint object

  std::map<uint32_t, cp_type> R; // constraint set: SL id -> c

  cp_type c = C.default_c();
  for(uint32_t i = 0; i < landmarks.size(); ++i) {
    if(R.find(p.L[landmarks[i]]) != R.end())
      continue; // already decided we're constrained wrt L[landmarks[i]]
    if(C.draw_constraint_params(feat, landmarks[i]->landmark, c))
      R[p.L[landmarks[i]]] = c; // we are constrained with respect to this SL
  }

  landmarks.push_back(lm); // add new landmark to the working set of landmarks

  if(R.size() == 0) // no constraints on the new landmark
    return;

  // get a set of all the landmarks constrained wrt the new one
  std::vector<map_type::iterator> Rlm;
  Rlm.push_back(lm);
  for(uint32_t i = 0; i < landmarks.size(); ++i)
    if(R.find(p.L[landmarks[i]]) != R.end())
      Rlm.push_back(landmarks[i]);

  // compute the ML estimate of the constrained parameters
  Vecd ml_mean, uncon_cp_mean;
  Matd ml_cov(C.constrained_params_size(), C.constrained_params_size(), vl_0), uncon_cp_cov;
  feature_type uncon;
  C.start_ml_mean(ml_mean);
  for(uint32_t i = 0; i < Rlm.size(); ++i) {
    uncon = merge(p.backup[Rlm[i]], p.accum[Rlm[i]]); // compute unconstrained state estimate
    C.get_constrained_params(uncon, uncon_cp_mean, uncon_cp_cov);
    uncon_cp_cov = inv(uncon_cp_cov);
    ml_cov = ml_cov + uncon_cp_cov;  // inv(P) = inv(P)+inv(uncon_cp_cov);
    C.h(R[p.L[Rlm[i]]], uncon_cp_mean); // apply constraint parameters

    // we do this in the constraint object in case there are special
    // considerations when computing the mean.  for example, the
    // rectilinearity constraint is applied to the angles of line
    // landmarks.  averaging angles directly doesn't work so we
    // average the sines/cosines and then compute atan2 at the end of
    // the averaging.
    C.add_to_ml_mean(ml_mean, uncon_cp_mean, uncon_cp_cov);
  }
  C.end_ml_mean(ml_mean);
  ml_mean = ml_cov * ml_mean; // inv(P)*mean
  ml_cov = inv(ml_cov);

  // condition all the constrained landmarks on the same values for
  // the constrained parameters
  for(uint32_t i = 0; i < Rlm.size(); ++i) {
    Rlm[i]->landmark = p.backup[Rlm[i]]; // "rewind" state to pre-particlized version
    // should set the landmark to have g(c,ml_mean) for the
    // constrained params, with zero covariance, and the remaining
    // parameters are conditioned on this
    C.conditionalize(Rlm[i]->landmark, R[p.L[Rlm[i]]], ml_mean);
    // "replay" measurements since particlization
    Rlm[i]->landmark = merge(Rlm[i]->landmark, p.accum[Rlm[i]]);
  }

  // remove old superlandmarks in R, and combine them all into a
  // single superlandmark; we do this at the end for bookkeeping
  // reasons although it happens near the beginning in the wafr06
  // pseudocode
  for(uint32_t i = 0; i < Rlm.size(); ++i) {
    p.L_size[p.L[landmarks[i]]] = 0;
    p.L[landmarks[i]] = p.L[lm];
    p.L_size[p.L[lm]]++;
  }
}

// update a landmark and take care of accumulator/backup-state
// bookkeeping
void slam_rbpf_constrained::update
  (constrained_particle_t &p, map_type::iterator lm, feature_type &feat)
{
  lm->landmark = merge(lm->landmark, feat);  // update state/covariance
  if(p.L_size[p.L[lm]] > 1) // update measurement accumulator
    p.accum[lm] = merge(p.accum[lm], feat);
  else // update backup state/covariance
    p.backup[lm] = lm->landmark;
}

void slam_rbpf_constrained::add_features
  (const std::vector<feature_type> &features,
   const std::vector<uint32_t> &associations,
   const pose_t &ref_frame)
{
  // most of this is copied directly from rbpf.cpp, with minor
  // modifications

  // either we know all associations or none
  assert(associations.size() == 0 || associations.size() == features.size());

  double W = 0, mahalanobis_dist = 0;
  uint32_t i, f;
  std::vector<map_type::iterator> in_range, all_lm;
  map_type::iterator assoc;
  bool add_new = false;
  feature_type feature;

  // iterate over all particles
  for(i = 0; i < particles.size(); ++i) {
    constrained_particle_t &p = particles[i];

    all_lm.clear(); // (DIFFERENT FROM RBPF.CPP)

    // find in-range landmarks
    if(associations.size() == 0)
      p.map.in_range(p.pose, param::max_range, in_range);

    // process one feature at a time
    for(f = 0; f < features.size(); ++f) {
      // transform into the frame of this particle
      feature_to_particle_frame(features[f], ref_frame, feature, p.pose);

      // do data association
      if(associations.size() != 0) { // known data association
	assoc = p.map.find_id(associations[f]);
	if(assoc != 0) {
	  mahalanobis_dist = mahalanobis_distance(assoc->landmark, feature);
	  add_new = false;
	} else
	  add_new = true;
      } else if(in_range.size() > 0) { // unknown data association
	data_association(feature, in_range, assoc, mahalanobis_dist);
	add_new = (mahalanobis_dist >= param::da_ml_threshold); // are we too far away?
      } else
	add_new = true;

      if(!add_new) { // found an association
	p.w *= equality_likelihood_from_mahalanobis(assoc->landmark, feature, mahalanobis_dist);
	assoc = p.map.modify(assoc); // make sure we can touch this landmark
	update(p, assoc, feature); // update landmark (DIFFERENT FROM RBPF.CPP)
	++p.obs_count[assoc];
      } else { // add new landmark to map
	if(all_lm.size() == 0)
	  p.map.get_all_nodes(all_lm); // get a working set of all landmarks
	                               // for the insertion algorithm (DIFFERENT FROM RBPF.CPP)
	if(associations.size() != 0)
	  insert(p, all_lm, feature, associations[f]); // known landmark id (DIFFERENT FROM RBPF.CPP)
	else
	  insert(p, all_lm, feature); // (DIFFERENT FROM RBPF.CPP)
	p.w *= exp(-0.5*param::da_ml_threshold*param::da_ml_threshold); // update importance weight
      }
    }

    if(isnan(p.w))
      p.w = 0;

    W += p.w; // update total importance weight over all particles

    // remove landmarks we should have seen but didn't
    // FIXME this doesn't deal with accumulator/backup/etc. bookkeeping
    //    remove_spurious_landmarks(p.pose, p.map, in_range, all_points, p.obs_count);
  }

  // this is kind of a catastrophic failure, but do what we can
  if(W == 0) {
    for(i = 0; i < particles.size(); ++i)
      particles[i].w = 1.0;
    W = double(particles.size());
  }

  // normalize importance weights and compute "best" particle
  best = 0;
  for(i = 0; i < particles.size(); ++i) {
    particles[i].w /= W;
    if(particles[i].w > particles[best].w)
      best = i;
  }

  Neff = neff(particles);

  // check if we need to resample, and if so, do it
  if(Neff < param::resample_neff_percent * particles.size())
    resample(particles);

  if(param::gui) {
    // for rendering later
    frame = ref_frame;
    cur_features = features;
  }

  do_analysis();
}


//
// implementation of a rectilinearity constraint for line landmarks
//

bool rectilinearity::draw_constraint_params
  (const extent_feature &f1, const extent_feature &f2,
   param_type &c) const
{
  const double cvals[4] = {0, M_PI/2, M_PI, 3*M_PI/2};
  const double a = angular_distance(f1.t(), f2.t()); // mean of f1.t()-f2.t()

  // HACK: only bother with the "closest" constraint coefficient since
  // for the rectilinearity case the likelihoods of the others will be
  // vanishingly small (if the constraint should really be enforced)
  c = cvals[0];
  for(uint32_t i = 1; i < 4; ++i)
    if(fabs(angular_distance(a, c)) > fabs(angular_distance(a, cvals[i])))
      c = cvals[i];

  // compute the likelihood that c-delta <= f1.t()-f2.t() <= c+delta
  const double var = f1.cov[1][1] + f2.cov[1][1]; // variance of f1.t()-f2.t()
  // FIXME check this
  const double prob = gsl_cdf_gaussian_P(c + param::rectilinearity_delta - a, var) -
    gsl_cdf_gaussian_P(c - param::rectilinearity_delta - a, var);

  // apply the constraint with probability proportional to the
  // likelihood
  return (drand48() < prob);
}

void rectilinearity::get_constrained_params
  (const extent_feature &f, Vecd &mu, Matd &cov) const
{
  mu[0] = f.t();
  cov[0][0] = f.cov[1][1];
}

void rectilinearity::h(const param_type &c, Vecd &mu) const
{
  mu[0] += c;
}

void rectilinearity::start_ml_mean(Vecd &mu) const
{
  mu = Vecd(2, vl_0);
}

// ucov is the inverse of the covariance of mu
void rectilinearity::add_to_ml_mean
  (Vecd &mu, const Vecd &umu, const Matd &ucov) const
{
  mu[0] += sin(umu[0]) * ucov[0][0];
  mu[1] += cos(umu[0]) * ucov[0][0];
}

void rectilinearity::end_ml_mean(Vecd &mu) const
{
  double a = atan2(mu[0], mu[1]);
  mu = Vecd(1, a);
}

void rectilinearity::conditionalize
  (extent_feature &f, const param_type &c, const Vecd &ml_mean) const
{
  const double g = clamp_angle(ml_mean[0] - c); // g(c,ml_mean)
  // conditional mean: r = r + rt*(1/tt)*(g(c,ml_mean) - t)
  f.r() += (f.cov[0][1] / f.cov[1][1]) * (angular_distance(g, f.t()));
  // conditional covariance: P_np = P_np - P_pnp*P_p^(-1)*P_pnp^T
  f.cov[0][0] -= (f.cov[0][1] / f.cov[1][1]) * f.cov[0][1];
  // fix constrained variable:
  f.t() = g;
  f.cov[0][1] = f.cov[1][0] = f.cov[1][1] = 0.0;

  /// FIXME FIXME: update extent
}
