/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

// detect landmarks that should have been observed but that weren't,
// and remove them; based on straightforward, brute-force "raytracing"
// of scan lines

// an alternative idea is to do this stochastically, i.e., select
// landmarks and/or scan lines at random, to save on processing

#include "slam.hpp"
#include "rbpf.hpp"

// assumes points has been transformed to the world frame as though
// the scan was taken from p.pose

// only tests landmarks for which observed != true

void remove_spurious_landmarks
  (const pose_t &p, map_t<extent_feature> &m,
   std::vector<map_t<extent_feature>::iterator> &in_range,
   const cartesian_scan_t &points,
   std::map<map_t<extent_feature>::iterator, int32_t> &obs_count)
{
  if(param::spurious_discard_count == 0)
    return;

  uint32_t i, j, hit_count;
  point_t fp1, fp2;
  double den, u;
  for(i = 0; i < in_range.size(); ++i) {
    const extent_feature &f = in_range[i]->landmark;
    hit_count = 0;
    fp1 = u_to_point(f, f.umin);
    fp2 = u_to_point(f, f.umax);
    for(j = 0; j < points.size(); ++j) {
      const point_t &sp = points[j];
      den = (fp2.y()-fp1.y())*(sp.x()-p.x()) - (fp2.x()-fp1.x())*(sp.y()-p.y());
      u = ((fp2.x()-fp1.x())*(p.y()-fp1.y()) - (fp2.y()-fp1.y())*(p.x()-fp1.x())) / den;
      if(u >= 0 && u <= 1) {
	u = ((sp.x()-p.x())*(p.y()-fp1.y()) - (sp.y()-p.y())*(p.x()-fp1.x())) / den;
	if(u >= 0 && u <= 1) { // the two line segments intersect!
	  ++hit_count;
	  if(hit_count > param::le_min_points)
	    break;
	}
      }
    }
    if(hit_count > param::le_min_points) {
      --obs_count[in_range[i]]; // should have seen it!
      if(obs_count[in_range[i]] < -int(param::spurious_discard_count))
	m.erase((map_t<extent_feature>::iterator &)in_range[i]); // kind of a hack to avoid templates
    }
  }
}

void remove_spurious_landmarks
  (const pose_t &p, map_t<point_feature> &m,
   std::vector<map_t<point_feature>::iterator> &in_range,
   const cartesian_scan_t &points,
   std::map<map_t<point_feature>::iterator, int32_t> &obs_count)
{
  if(param::spurious_discard_count == 0)
    return;

  uint32_t i;
  double a;
  for(i = 0; i < in_range.size(); ++i) {
    const point_feature &f = in_range[i]->landmark;
    // is the feature in our field of view?
    a = clamp_angle(atan2(f.y() - p.y(), f.x() - p.x()) - p.t());
    if(a > param::scan_min && a < param::scan_max) {
      --obs_count[in_range[i]];
      if(obs_count[in_range[i]] < -int(param::spurious_discard_count))
	m.erase((map_t<point_feature>::iterator &)in_range[i]);
    }
  }
}
