#!/usr/bin/perl

$M = 50;   # number of trials
$N = 500; # number of particles

# FIXME: make these part of the array below
$sim_map = "sim.lmk";
$sim_feat = "sim.feat";
$sim_log = "sim-odom"; # assumes a .phist and a .dat exist
$sim_truth = "sim-truth.phist";
$real_feat = "../slam/data/sets/car-park.feat";
$real_log = "../slam/data/sets/car-park";
$real_conf = "car-park.conf";
$slamexec = "./slam";

# arguments:
#  directory with input/ subdir
#  output subdir for the experiment
#  slam algorithm (fs1,fs2,rough,block)
#  resampling algorithm (ran,res,gen)
#  generalized resampling alpha (only for gen resampler)
#  lag time (only for roughening or block sisr)
@experiments = (

# 		["results/sim-4", "fs2-ran", "fs2", "ran", 0.0, 0, 1],
# 		["results/sim-4", "fs2-res", "fs2", "res", 0.0, 0, 1],
# 		["results/sim-4", "fs2-gen-a0.2", "fs2", "gen", 0.2, 0, 1],
# 		["results/sim-4", "fs2-gen-a0.5", "fs2", "gen", 0.5, 0, 1],
# 		["results/sim-4", "fs2-gen-a0.8", "fs2", "gen", 0.8, 0, 1],
# 		["results/sim-4", "rough-ran-l1", "rough", "ran", 0.0, 1, 1],
# 		["results/sim-4", "rough-ran-l5", "rough", "ran", 0.0, 5, 1],
# 		["results/sim-4", "rough-ran-l10", "rough", "ran", 0.0, 10, 1],
# 		["results/sim-4", "block-ran-l1", "block", "ran", 0.0, 1, 1],
# 		["results/sim-4", "block-ran-l5", "block", "ran", 0.0, 5, 1],
# 		["results/sim-4", "block-ran-l10", "block", "ran", 0.0, 10, 1],

# 		["results/sim-6", "fs2-ran", "fs2", "ran", 0.0, 0, 1],
# 		["results/sim-6", "fs2-res", "fs2", "res", 0.0, 0, 1],
# 		["results/sim-6", "fs2-gen-a0.2", "fs2", "gen", 0.2, 0, 1],
# 		["results/sim-6", "fs2-gen-a0.5", "fs2", "gen", 0.5, 0, 1],
# 		["results/sim-6", "fs2-gen-a0.8", "fs2", "gen", 0.8, 0, 1],
# 		["results/sim-6", "rough-ran-l1", "rough", "ran", 0.0, 1, 1],
# 		["results/sim-6", "rough-ran-l5", "rough", "ran", 0.0, 5, 1],
# 		["results/sim-6", "rough-ran-l10", "rough", "ran", 0.0, 10, 1],
# 		["results/sim-6", "block-ran-l1", "block", "ran", 0.0, 1, 1],
# 		["results/sim-6", "block-ran-l5", "block", "ran", 0.0, 5, 1],
# 		["results/sim-6", "block-ran-l10", "block", "ran", 0.0, 10, 1],

		["results/car-park", "fs2-ran", "fs2", "ran", 0.0, 0, 0],
#		["results/car-park", "fs2-res", "fs2", "res", 0.0, 0, 0],
#		["results/car-park", "fs2-gen-a0.2", "fs2", "gen", 0.2, 0, 0],
#		["results/car-park", "fs2-gen-a0.5", "fs2", "gen", 0.5, 0, 0],
#		["results/car-park", "fs2-gen-a0.8", "fs2", "gen", 0.8, 0, 0],
		["results/car-park", "rough-ran-l1", "rough", "ran", 0.0, 1, 0],
		["results/car-park", "rough-ran-l5", "rough", "ran", 0.0, 5, 0],
		["results/car-park", "rough-ran-l10", "rough", "ran", 0.0, 10, 0],
		["results/car-park", "block-ran-l1", "block", "ran", 0.0, 1, 0],
		["results/car-park", "block-ran-l5", "block", "ran", 0.0, 5, 0],
		["results/car-park", "block-ran-l10", "block", "ran", 0.0, 10, 0],

		);


foreach $exp (@experiments) {
    $inprefix = "$$exp[0]/input";
    $prefix = "$$exp[0]/$$exp[1]/mc-";
    $slam = $$exp[2];
    $resample = $$exp[3];
    $alpha = $$exp[4];
    $lag = $$exp[5];
    $is_sim = $$exp[6];

    $dir = "$$exp[0]/$$exp[1]";
    if(! -d $dir) { `mkdir $dir`; }

    experiment($inprefix, $prefix, $slam, $resample, $alpha, $lag, $is_sim);
}


sub experiment {
    my ($inprefix, $prefix, $slam, $resample, $alpha, $lag, $is_sim) = @_;

    print "Generating $prefix files: ";

    for($i=0; $i<$M; ++$i) {
	$nprefix = "$prefix$i";

	if(-f "$nprefix.mcphist") { print "."; next; }

	if($is_sim) {
	    `./sim --err-range 0.05 --err-bearing 0.005 -m $inprefix/$map -t $inprefix/$truth --nogui`;
	    `utils/phist2dat.pl $sim_log.phist >$sim_log.dat`;
	    $origin = `head -n 1 $sim_log.phist`;
	    ($ox,$oy,$ot) = split /\s+/, $origin;
	    `$slamexec -n $N --features-file $sim_feat $sim_log.dat --origin-x $ox --origin-y $oy --origin-theta $ot -a $slam -r $resample --resample-alpha $alpha -l $lag >/dev/null`;
	    `mv sim-odom.phist $nprefix.odom.phist`;
	    `mv sim.feat $nprefix.feat`;
	} else {
	    `$slamexec -c $real_conf -n $N --features $real_feat $real_log.dat -a $slam -r $resample --resample-alpha $alpha -l $lag >/dev/null`;
	}

	`mv output.mcphist $nprefix.mcphist`;
	`mv output.mcfull $nprefix.mcfull`;
	`mv output.neff $nprefix.neff`;
	`mv output.lmk $nprefix.lmk`;
	`mv output.unique $nprefix.unique`;
	`mv output.perf $nprefix.perf`;
#	removeline("$nprefix.mcfull");

	print ".";
    }

    print "\n";
}

sub removeline
{
    my ($f) = @_;
    if(! -f $f) { return; }
    open FILE,"<$f";
    open OUT,">foo";
    <FILE>;
    while(<FILE>) { print OUT; }
    close FILE;
    close OUT;
    `mv foo $f`;
}
